#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


from oc_meta.lib.file_manager import pathoo, get_data, write_csv, sort_files
from oc_meta.lib.master_of_regex import comma_and_spaces, name_and_ids, semicolon_in_people_field
from oc_meta.scripts.creator import Creator
from typing import Dict, List
from tqdm import tqdm
import os
import psutil
import re


FORBIDDEN_IDS = {'issn:0000-0000'}
VENUES = {'archival-document', 'book', 'book-part', 'book-section', 'book-series', 'book-set', 'edited-book', 'journal', 'journal-volume', 'journal-issue', 'monograph', 'proceedings-series', 'proceedings', 'reference-book', 'report-series', 'standard-series'}

def prepare_relevant_items(csv_dir:str, output_dir:str, items_per_file:int, verbose:bool) -> None:
    '''
    This function receives an input folder containing CSVs formatted for Meta. 
    It output other CSVs, including deduplicated items only. 
    You can specify how many items to insert in each output file.

    :params csv_dir: the path to the folder containing the input CSV files
    :type csv_dir: str
    :params output_dir: the location of the folder to save to output file
    :type output_dir: str
    :params items_per_file: an integer to specify how many rows to insert in each output file
    :type items_per_file: int
    :params verbose: if True, show a loading bar, elapsed, and estimated time
    :type verbose: bool
    :returns: None -- This function returns None and saves the output CSV files in the `output_dir` folder
    '''
    files = [os.path.join(csv_dir, file) for file in sort_files(os.listdir(csv_dir)) if file.endswith('.csv')]
    pbar = tqdm(total=len(files)) if verbose else None
    pathoo(output_dir)
    ids_found = set()
    venues_found = dict()
    duplicated_ids = dict()
    venues_by_id = dict()
    duplicated_venues = dict()
    publishers_found = set()
    publishers_by_id = dict()
    duplicated_publishers = dict()
    resp_agents_found = set()
    resp_agents_by_id = dict()
    duplicated_resp_agents = dict()
    # Look for all venues, responsible agents, and publishers
    for file in files:
        data = get_data(file)
        _get_duplicated_ids(data=data, ids_found=ids_found, items_by_id=duplicated_ids)
        _get_relevant_venues(data=data, ids_found=venues_found, items_by_id=venues_by_id, duplicated_items=duplicated_venues)
        _get_publishers(data=data, ids_found=publishers_found, items_by_id=publishers_by_id, duplicated_items=duplicated_publishers)
        _get_resp_agents(data=data, ids_found=resp_agents_found, items_by_id=resp_agents_by_id, duplicated_items=duplicated_resp_agents)
        pbar.update() if verbose else None
    pbar.close() if verbose else None
    if verbose:
        print('[INFO:prepare_multiprocess] Enriching the duplicated bibliographic resources found')
        pbar = tqdm(total=len(files))
    for file in files:
        data = get_data(file)
        _enrich_duplicated_ids_found(data, duplicated_ids)
        pbar.update() if verbose else None
    pbar.close() if verbose else None
    ids_merged = _do_collective_merge(duplicated_ids, duplicated_ids)
    venues_merged = _do_collective_merge(venues_by_id, duplicated_venues)
    publishers_merged = _do_collective_merge(publishers_by_id, duplicated_publishers)
    resp_agents_merged = _do_collective_merge(resp_agents_by_id, duplicated_resp_agents)
    fieldnames = ['id', 'title', 'author', 'pub_date', 'venue', 'volume', 'issue', 'page', 'type', 'publisher', 'editor']
    __save_relevant_venues(venues_merged, items_per_file, output_dir, fieldnames)
    __save_ids(ids_merged, items_per_file, output_dir, fieldnames)
    __save_responsible_agents(resp_agents_merged, items_per_file, output_dir, fieldnames, ('people', 'author'))
    __save_responsible_agents(publishers_merged, items_per_file, output_dir, fieldnames, ('publishers', 'publisher'))

def _get_duplicated_ids(data:List[dict], ids_found:set, items_by_id:Dict[str, dict]) -> None:
    cur_file_ids = set()
    for row in data:
        ids_list = row['id'].split()
        if any(id in ids_found and (id not in cur_file_ids or id in items_by_id) for id in ids_list):
            for id in ids_list:
                items_by_id.setdefault(id, {'others': set()})
                items_by_id[id]['others'].update({other for other in ids_list if other != id})
                for field in ['title', 'author', 'pub_date', 'venue', 'volume', 'issue', 'page', 'type', 'publisher', 'editor']:
                    if field in items_by_id[id]:
                        if len(row[field]) < len(items_by_id[id][field]):
                            continue
                    items_by_id[id][field] = row[field]
        cur_file_ids.update(set(ids_list))   
        ids_found.update(set(ids_list))

def _enrich_duplicated_ids_found(data:List[dict], items_by_id:Dict[str, dict]) -> None:
    for row in data:
        br_ids = row['id'].split()
        for br_id in br_ids:
            if br_id in items_by_id:
                all_ids = __find_all_ids_by_key(items_by_id, br_id)
                for field, value in row.items():
                    if field != 'id':
                        for all_id in all_ids:
                            if len(value) > len(items_by_id[all_id][field]):
                                items_by_id[all_id][field] = value

def _get_relevant_venues(data:List[dict], ids_found:dict, items_by_id:Dict[str, dict], duplicated_items:Dict[str, dict]) -> None:
    cur_file_ids = dict()
    for row in data:
        venue = row['venue']
        venues = list()
        if row['type'] in VENUES:
            venues.append((row['title'], row['id'], row['type']))
        if venue:
            full_name_and_ids = re.search(name_and_ids, venue)
            name = full_name_and_ids.group(1) if full_name_and_ids else venue
            ids = full_name_and_ids.group(2) if full_name_and_ids else None
            if ids:
                try:
                    br_type = Creator.get_venue_type(row['type'], ids.split())
                except UnboundLocalError:
                    print(f"[INFO:prepare_multiprocess] I found the venue {row['venue']} for the resource of type {row['type']}, but I don't know how to handle it")
                    raise UnboundLocalError
                venues.append((name, ids, br_type))
        for venue_tuple in venues:
            name, ids, br_type = venue_tuple
            ids_list = [identifier for identifier in ids.split() if identifier not in FORBIDDEN_IDS]
            if any(id in ids_found and (id not in cur_file_ids or id in duplicated_items) for id in ids_list):
                for id in ids_list:
                    duplicated_items.setdefault(id, {'others': set(), 'name': name, 'type': br_type, 'volume': dict(), 'issue': set()})
                    duplicated_items[id]['others'].update({other for other in ids_list if other != id})
            for id in ids_list:
                items_by_id.setdefault(id, {'others': set(), 'name': name, 'type': br_type, 'volume': dict(), 'issue': set()})
                items_by_id[id]['others'].update({other for other in ids_list if other != id})
            for id in ids_list:
                ids_found.setdefault(id, {'volumes': dict(), 'issues': set()})
                cur_file_ids.setdefault(id, {'volumes': dict(), 'issues': set()})
            volume = row['volume']
            issue = row['issue']
            if volume:
                if any(volume in ids_found[id]['volumes'] and volume not in cur_file_ids[id]['volumes'] for id in ids_list):
                    for id in ids_list:
                        duplicated_items[id]['volume'].setdefault(volume, set())
                for id in ids_list:
                    ids_found[id]['volumes'].setdefault(volume, set())
                    cur_file_ids[id]['volumes'].setdefault(volume, set())                        
                if issue:
                    if any(issue in ids_found[id]['volumes'][volume] and issue not in cur_file_ids[id]['volumes'][volume] for id in ids_list):
                        for id in ids_list:
                            duplicated_items[id]['volume'].setdefault(volume, set())
                            duplicated_items[id]['volume'][volume].add(issue)
                    for id in ids_list:
                        cur_file_ids[id]['volumes'][volume].add(issue)
                        ids_found[id]['volumes'][volume].add(issue)
            elif not volume and issue:
                if any(issue in ids_found[id]['issues'] and issue not in cur_file_ids[id]['issues'] for id in ids_list):
                    for id in ids_list:
                        duplicated_items[id]['issue'].add(issue)
                for id in ids_list:
                    ids_found[id]['issues'].add(row['issue'])
                    cur_file_ids[id]['issues'].add(row['issue'])
                
def _get_publishers(data:List[dict], ids_found:set, items_by_id:Dict[str, dict], duplicated_items:Dict[str, dict]) -> None:
    cur_file_ids = set()
    for row in data:
        pub_name_and_ids = re.search(name_and_ids, row['publisher'])
        if pub_name_and_ids:
            pub_name = pub_name_and_ids.group(1)
            pub_ids = pub_name_and_ids.group(2).split()
            if any(id in ids_found and (id not in cur_file_ids or id in duplicated_items) for id in pub_ids):
                for pub_id in pub_ids:
                    duplicated_items.setdefault(pub_id, {'others': set(), 'name': pub_name, 'type': 'publisher'})
                    duplicated_items[pub_id]['others'].update({other for other in pub_ids if other != pub_id})
            for pub_id in pub_ids:
                items_by_id.setdefault(pub_id, {'others': set(), 'name': pub_name, 'type': 'publisher'})
                items_by_id[pub_id]['others'].update({other for other in pub_ids if other != pub_id})
            cur_file_ids.update(set(pub_ids))   
            ids_found.update(set(pub_ids))

def _get_resp_agents(data:List[dict], ids_found:set, items_by_id:Dict[str, Dict[str, set]], duplicated_items:Dict[str, dict]) -> None:
    cur_file_ids = set()
    for row in data:
        for field in {'author', 'editor'}:
            if row[field]:
                resp_agents = re.split(semicolon_in_people_field, row[field])
                for resp_agent in resp_agents:
                    full_name_and_ids = re.search(name_and_ids, resp_agent)
                    name = full_name_and_ids.group(1) if full_name_and_ids else resp_agent
                    ids = full_name_and_ids.group(2) if full_name_and_ids else None
                    if ids:
                        ids_list = [identifier for identifier in ids.split() if identifier not in FORBIDDEN_IDS]
                        richest_name = _find_all_names(duplicated_items, ids_list, name)
                        if any(id in ids_found and (id not in cur_file_ids or id in duplicated_items) for id in ids_list):
                            for id in ids_list:
                                duplicated_items.setdefault(id, {'others': set(), 'type': 'author'})
                                duplicated_items[id]['name'] = richest_name
                                duplicated_items[id]['others'].update({other for other in ids_list if other != id})
                        for id in ids_list:
                            items_by_id.setdefault(id, {'others': set(), 'type': 'author'})
                            items_by_id[id]['name'] = richest_name
                            items_by_id[id]['others'].update({other for other in ids_list if other != id})
                        cur_file_ids.update(set(ids_list))   
                        ids_found.update(set(ids_list))

def _find_all_names(items_by_id:Dict[str, Dict[str, set]], ids_list:list, cur_name:str) -> str:
    all_names = {cur_name}
    for id in ids_list:
        if id in items_by_id:
            all_names.add(items_by_id[id]['name'])
    all_names_parsed = set()
    for name in all_names:
        if ',' in name:
            split_name = re.split(comma_and_spaces, name)
            first_name = split_name[1].strip()
            surname = split_name[0].strip()
            all_names_parsed.add((surname, first_name))
        else:
            all_names_parsed.add((name,))
    richest_first_name = ''
    richest_surname = ''
    for name in all_names_parsed:
        if name[0] > richest_surname:
            richest_surname = name[0]
        if len(name) == 2:
            if name[1] > richest_first_name:
                richest_first_name = name[1]
    return f'{richest_surname}, {richest_first_name}'.strip()

def _do_collective_merge(items_by_id:dict, duplicated_items:Dict[str, dict]) -> dict:
    merged_by_key:Dict[str, Dict[str, set]] = dict()
    ids_checked = set()
    for id, data in duplicated_items.items():
        if id not in ids_checked:
            all_vi = None
            ids_found = {id}
            ids_to_be_checked = {id}
            ids_to_be_checked.update(data['others'])
            for id_to_be_checked in ids_to_be_checked:
                if id_to_be_checked not in ids_checked:
                    output_ids = __find_all_ids_by_key(items_by_id, key=id_to_be_checked)
                    ids_checked.update(output_ids)
                    ids_found.update(output_ids)
            all_other_ids = {item for item in ids_found if item != id}
            ids_checked.update(ids_found)
            if 'volume' in data and 'issue' in data:
                if isinstance(data['volume'], dict):
                    all_vi = __find_all_vi(items_by_id=duplicated_items, all_ids=ids_found)
            if data['type'] == 'author':
                richest_name = _find_all_names(items_by_id, ids_found, data['name'])
                data['name'] = richest_name
            merged_by_key[id] = {k:v if not k == 'others' else all_other_ids for k,v in data.items()}
            if all_vi:
                merged_by_key[id]['volume'] = all_vi['volume']
                merged_by_key[id]['issue'] = all_vi['issue']
    del items_by_id
    del duplicated_items
    return merged_by_key

def __find_all_ids_by_key(items_by_id:dict, key:str):
    visited_items = set()
    items_to_visit = {item for item in items_by_id[key]['others']}
    while items_to_visit:
        for item in set(items_to_visit):
            if item not in visited_items:
                visited_items.add(item)
                items_to_visit.update({item for item in items_by_id[item]['others'] if item not in visited_items})
            items_to_visit.remove(item)
    return visited_items

def __find_all_vi(items_by_id:dict, all_ids:set) -> dict:
    all_vi = {'volume': dict(), 'issue': set()}
    for id in all_ids:
        if id in items_by_id:
            for volume, volume_issues in items_by_id[id]['volume'].items():
                all_vi['volume'].setdefault(volume, set()).update(volume_issues)
            for venue_issues in items_by_id[id]['issue']:
                all_vi['issue'].add(venue_issues)
    return all_vi

def __save_relevant_venues(items_by_id:dict, items_per_file:int, output_dir:str, fieldnames:list):
    output_dir = os.path.join(output_dir, 'venues')
    rows = list()
    counter = 0
    for item_id, data in items_by_id.items():
        item_type = data['type']
        row = dict()
        name, ids = __get_name_and_ids(item_id, data)
        row['id'] = ids
        row['title'] = name
        row['type'] = item_type
        for volume, volume_issues in data['volume'].items():
            volume_row = dict()
            volume_row['volume'] = volume
            volume_row['venue'] = f'{name} [{ids}]'
            if volume_issues:
                volume_row['type'] = 'journal issue'
                for volume_issue in volume_issues:
                    volume_issue_row = dict(volume_row)
                    volume_issue_row['issue'] = volume_issue
                    rows.append(volume_issue_row)
            else:
                volume_row['type'] = 'journal volume'
                rows.append(volume_row)
        for venue_issue in data['issue']:
            issue_row = dict()
            issue_row['venue'] = f'{name} [{ids}]'
            issue_row['issue'] = venue_issue
            issue_row['type'] = 'journal issue'
            rows.append(issue_row)
        if not data['volume'] and not data['issue']:
            rows.append(row)
        if len(rows) >= items_per_file:
            output_path = os.path.join(output_dir, f"{counter}.csv")
            write_csv(output_path, rows, fieldnames)
            rows = list()
            counter += 1
    output_path = os.path.join(output_dir, f"{counter}.csv")
    write_csv(output_path, rows, fieldnames)

def __save_responsible_agents(items_by_id:dict, items_per_file:int, output_dir:str, fieldnames:list, datatype:tuple):
    folder = datatype[0]
    field = datatype[1]
    output_dir = os.path.join(output_dir, folder)
    rows = list()
    chunks = int(items_per_file)
    saved_chunks = 0
    output_length = len(items_by_id)
    for item_id, data in items_by_id.items():
        name, ids = __get_name_and_ids(item_id, data)
        rows.append({field: f'{name} [{ids}]'})
        rows, saved_chunks = __store_data(rows, output_length, chunks, saved_chunks, output_dir, fieldnames)

def __save_ids(items_by_id:dict, items_per_file:int, output_dir:str, fieldnames:list):
    output_dir = os.path.join(output_dir, 'ids')
    rows = list()
    chunks = int(items_per_file)
    saved_chunks = 0
    output_length = len(items_by_id)
    for item_id, data in items_by_id.items():
        ids_list = list(data['others'])
        ids_list.append(item_id)
        ids = ' '.join(ids_list)
        output_row = {'id': ids}
        for field in [f for f in fieldnames if f != 'id']:
            output_row[field] = data[field]
        rows.append(output_row)
        rows, saved_chunks = __store_data(rows, output_length, chunks, saved_chunks, output_dir, fieldnames)

def __store_data(rows:list, output_length:int, chunks:int, saved_chunks:int, output_dir:str, fieldnames:str) -> list:
    data_about_to_end = (output_length - saved_chunks) < chunks and (output_length - saved_chunks) == len(rows)
    if len(rows) == chunks or data_about_to_end:
        saved_chunks = saved_chunks + chunks if not data_about_to_end else output_length
        filename = f'{str(saved_chunks)}.csv'
        output_path = os.path.join(output_dir, filename)
        write_csv(path=output_path, datalist=rows, fieldnames=fieldnames)
        rows = list()
    return rows, saved_chunks

def __get_name_and_ids(item_id:str, data:dict):
    ids_list = list(data['others'])
    ids_list.append(item_id)
    name = data['name'] if 'name' in data else ''
    ids = ' '.join(ids_list)
    return name, ids

def split_csvs_in_chunks(csv_dir:str, output_dir:str, chunk_size:int, verbose:bool=False) -> None:
    '''
    This function splits all CSVs in a folder in smaller CSVs having a specified number of rows.
    Moreover, this function tries, where possible, to keep in a single file the bibliographic resources contained in the same venue. 
    For this reason, the final rows number could be slightly over the specified one.

    :params csv_dir: the path to the folder containing the input CSV files
    :type csv_dir: str
    :params output_dir: the location of the folder to save to output files
    :type output_dir: str
    :params chunk_size: an integer to specify how many rows to insert in each output file
    :type chunk_size: int
    :params verbose: if True, show a loading bar, elapsed, and estimated time
    :type verbose: bool
    :returns: None -- This function returns None and saves the output CSV files in the `output_dir` folder
    '''
    files = [os.path.join(csv_dir, file) for file in sort_files(os.listdir(csv_dir)) if file.endswith('.csv')]
    pid = psutil.Process(os.getpid())
    venues_occurrences = __index_all_venues(files, verbose)
    __split_csvs_by_venues(files, venues_occurrences, output_dir, pid, verbose)
    __split_in_chunks(output_dir, chunk_size, verbose)

def __index_all_venues(files:list, verbose:bool) -> dict:
    if verbose:
        print('[INFO:prepare_multiprocess] Scanning venues')
        pbar = tqdm(total=len(files))
    venues_occurrences = dict()
    for file in files:
        data = get_data(file)
        for row in data:
            venues = list()
            if row['type'] in VENUES:
                venues.append(row['id'].split())
            venue_and_ids = re.search(name_and_ids, row['venue'])
            if venue_and_ids:
                ids = venue_and_ids.group(2).split()
                venues.append(ids)
            for venue_ids in venues:
                for venue_id in venue_ids:
                    venues_occurrences.setdefault(venue_id, {'others': set()})
                    venues_occurrences[venue_id]['others'].update({other for other in venue_ids if other != venue_id})
        pbar.update() if verbose else None
    pbar.close() if verbose else None
    return venues_occurrences

def __split_csvs_by_venues(files:list, venues_occurrences:dict, output_dir:str, pid:psutil.Process, verbose:bool):
    pathoo(output_dir)
    if verbose:
        print('[INFO:prepare_multiprocess] Splitting CSVs by venue')
        pbar = tqdm(total=len(files))
    chunk_venues = dict()
    chunk_no_venues = dict()
    existing_files = set()
    no_venues_outdata = list()
    counter = 0
    for file in files:
        data = get_data(file)
        for row in data:
            venues = list()
            if row['type'] in VENUES:
                venues.append(row['id'].split())
            venue_and_ids = re.search(name_and_ids, row['venue'])
            if venue_and_ids:
                ids = venue_and_ids.group(2).split()
                venues.append(ids)
            if venues:
                output_filepath = None
                for venue_ids in venues:
                    all_ids:list = venue_ids
                    all_ids.extend(__find_all_ids_by_key(venues_occurrences, key=all_ids[0]))
                    for any_id in all_ids:
                        filename = any_id.replace(':', '').replace('/', '').replace('\\', '')
                        if os.path.join(output_dir, f'{filename}.csv') in existing_files:
                            output_filepath = os.path.join(output_dir, f'{filename}.csv')
                filename = all_ids[0].replace(':', '').replace('/', '').replace('\\', '')
                output_filepath = os.path.join(output_dir, f'{filename}.csv') if not output_filepath else output_filepath
                chunk_venues.setdefault(output_filepath, list()).append(row)
                chunk_venues = __dump_if_chunk_size(chunk_venues, existing_files, pid)
            elif not venues:
                no_venues_outdata.append(row)
                if len(no_venues_outdata) == 1000:
                    no_venues_filepath = os.path.join(output_dir, f'no_venues_{counter}.csv')
                    chunk_no_venues[no_venues_filepath] = no_venues_outdata
                    counter += 1
                    no_venues_outdata = list()
                chunk_no_venues = __dump_if_chunk_size(chunk_no_venues, existing_files, pid)
        pbar.update() if verbose else None
    pbar.close() if verbose else None
    if no_venues_outdata:
        no_venues_filepath = os.path.join(output_dir, f'no_venues_{counter}.csv')
        chunk_no_venues[no_venues_filepath] = no_venues_outdata
    for chunk in [chunk_venues, chunk_no_venues]:
        for filepath, dump in chunk.items():
            all_data = get_data(filepath) if os.path.exists(filepath) else list()
            all_data.extend(dump)
            write_csv(filepath, all_data)
        del chunk
    del venues_occurrences

def __split_in_chunks(output_dir:str, chunk_size:int, verbose:bool):
    files = os.listdir(output_dir)
    if verbose:
        print('[INFO:prepare_multiprocess] Splitting CSVs in chunks')
        pbar = tqdm(total=len(files))
    even_chunk = list()
    counter = 0
    for file in files:
        filepath = os.path.join(output_dir, file)
        data = get_data(filepath)
        len_data = len(data)
        if len_data > chunk_size:
            while len_data > chunk_size:
                write_csv(os.path.join(output_dir, f'{counter}.csv'), data[:chunk_size])
                counter += 1
                del data[:chunk_size]
                len_data = len(data)
            even_chunk.extend(data)
            if len(even_chunk) >= chunk_size:
                write_csv(os.path.join(output_dir, f'{counter}.csv'), even_chunk)
                counter += 1
                even_chunk = list()
        elif len_data <= chunk_size:
            even_chunk.extend(data)
            if len(even_chunk) >= chunk_size:
                write_csv(os.path.join(output_dir, f'{counter}.csv'), even_chunk)
                counter += 1
                even_chunk = list()
        os.remove(filepath)
        pbar.update() if verbose else None
    pbar.close() if verbose else None
    if even_chunk:
        write_csv(os.path.join(output_dir, f'{counter}.csv'), even_chunk)

def __dump_if_chunk_size(chunk:dict, existing_files:set, pid:psutil.Process) -> dict:
    memory_used = pid.memory_info().rss / (1024.0 ** 3)
    if memory_used > 10:
        for filepath, dump in chunk.items():
            all_data = get_data(filepath) if os.path.exists(filepath) else list()
            all_data.extend(dump)
            write_csv(filepath, all_data)
            existing_files.add(filepath)
        return dict()
    return chunk

def __find_all_ids_by_key(items_by_id:dict, key:str):
    visited_items = set()
    items_to_visit = {item for item in items_by_id[key]['others']}
    while items_to_visit:
        for item in set(items_to_visit):
            if item not in visited_items:
                visited_items.add(item)
                items_to_visit.update({item for item in items_by_id[item]['others'] if item not in visited_items})
            items_to_visit.remove(item)
    return visited_items



import re
import html
from dateutil.parser import parse
from datetime import datetime
from oc_meta.lib.master_of_regex import *
from oc_meta.lib.id_manager.issnmanager import ISSNManager
from oc_meta.lib.id_manager.isbnmanager import ISBNManager
from oc_meta.lib.id_manager.doimanager import DOIManager
from oc_meta.lib.id_manager.orcidmanager import ORCIDManager
from oc_meta.lib.csvmanager import CSVManager
from oc_meta.lib.master_of_regex import volumes_valid_patterns, issues_valid_patterns, invalid_vi_patterns
from typing import Union, Tuple

class Cleaner:
    def __init__(self, string:str):
        '''
        :params string: the string to be cleaned.
        :type string: str
        '''
        self.string = string

    def normalize_hyphens(self) -> str:
        '''
        It replaces any hyphen, dash and minus sign with a hyphen-minus character.
        This is done for pages, IDs and dates.

        .. list-table:: Comparison between the various characters similar to hyphen-minus
            :widths: 25 25 50
            :header-rows: 1

            * - UTF-8
                - SIGN
                - NAME
            * - U+002D
                - -
                - Hyphen-minus
            * - U+00AD
                - ­
                - Soft hyphen
            * - U+06D4
                - ۔
                - Arabic Full Stop
            * - U+2010
                - ‐
                - Hyphen
            * - U+2011
                - −
                - Non-breaking Hyphen
            * - U+2012
                - –
                - Figure Dash
            * - U+2013
                - –
                - En-Dash
            * - U+2014
                - —
                - Em-Dash
            * - U+2043
                - ⁃
                - Hyphen Bullet
            * - U+2212
                - −
                - Minus Sign
            * - U+2796
                - ➖
                - Heavy Minus Sign
            * - U+2CBA
                - Ⲻ
                - Coptic Capital Letter Dialect-p Ni
            * - U+FE58
                - ﹘
                - Small Em Dash

        :returns: str -- the string with normalized hyphens
        '''
        string = self.string
        wrong_characters = {'\u00AD', '\u06D4', '\u2010', '\u2011', '\u2012', '\u2013', '\u2014', '\u2043', '\u2212', '\u2796', '\u2CBA', '\uFE58'}
        for c in wrong_characters:
            string = string.replace(c, '\u002D')
        if 'isbn:' in string:
            string.replace(u'\u002D', '')
        return string
    
    def normalize_spaces(self) -> str:
        '''
        It replaces any ambiguous spaces with a space.

        .. list-table:: List of the various characters similar to the space
            :widths: 25 25 50
            :header-rows: 1

            * - UTF-8
                - NAME
            * - U+0020
                - Space
            * - U+0009
                - Character Tabulation
            * - U+00A0
                - No-break space
            * - U+200B
                - Zero width space
            * - U+202F
                - Narrow no-break space
            * - U+2003
                - Em Space
            * - U+2005
                - Four-Per-Em Space
            * - U+2009
                - Thin Space

        :returns: str -- the string with normalized spaces
        '''
        string = self.string
        wrong_characters = {'\u0009', '\u00A0', '&nbsp;', '\u200B', '\u202F', '\u2003', '\u2005', '\u2009'}
        for c in wrong_characters:
            string = string.replace(c, '\u0020')
        return string

    def clean_title(self) -> str:
        '''
        Concerning titles of bibliographic resources ('venue' and 'title' columns), 
        every word in the title is capitalized except for those that have capitals within them 
        (probably acronyms, e.g. 'FaBiO and CiTO'). This exception, however, does not include entirely capitalized titles. 
        Finally, null characters and spaces are removed.

        :returns: str -- The cleaned title
        '''
        title = self.string
        if title.isupper():
            title = title.lower()
        words = title.split()
        for i, w in enumerate(words):
            if not any(x.isupper() for x in w):
                words[i] = w.title()
        new_title = ' '.join(words)
        return new_title

    def __date_parse_hack(self, date:str) -> datetime:
        dt = parse(date, default=datetime(2001, 1, 1))
        dt2 = parse(date, default=datetime(2002, 2, 2))

        if dt.year == dt2.year and dt.month == dt2.month and dt.day == dt2.day:
            clean_date = parse(date).strftime('%Y-%m-%d')
        elif dt.year == dt2.year and dt.month == dt2.month:
            clean_date = parse(date).strftime('%Y-%m')
        elif dt.year == dt2.year:
            clean_date = parse(date).strftime('%Y')
        else:
            clean_date = ''
        return clean_date

    def clean_date(self) -> str:
        '''
        It tries to parse a date-string into a datetime object, 
        considering both the validity of the format (YYYYY-MM-DD) and the value (e.g. 30 February is not a valid date). 
        For example, a date 2020-02-30 will become 2020-02, because the day is invalid. 
        On the other hand, 2020-27-12 will become 2020 since the day
        and month are invalid. 
        If the year is not valid (e.g.year >9999) data would be totally discarded.

        :returns: str -- The cleaned date or an empty string
        '''
        date = self.string
        try:
            date = self.__date_parse_hack(date)
        except ValueError:
            try:
                # e.g. 2021-12-17
                if len(date) == 10:
                    try:
                        # Maybe only the day is invalid, try year-month
                        new_date = date[:-3]
                        date = self.__date_parse_hack(new_date)
                    except ValueError:
                        try:
                            # Maybe only the month is invalid, try year
                            new_date = date[:-6]
                            date = self.__date_parse_hack(new_date)
                        except ValueError:
                            date = ''
                # e.g. 2021-12
                elif len(date) == 7:
                    # Maybe only the month is invalid, try year
                    try:
                        new_date = date[:-3]
                        date = self.__date_parse_hack(new_date)
                    except ValueError:
                        date = ''
                else:
                    date = ''
            except ValueError:
                date = ''
        return date

    def clean_name(self) -> str:
        '''
        The first letter of each element of the name is capitalized and superfluous spaces are removed.

        :returns: str -- The cleaned name
        '''
        name = self.string
        if ',' in name:
            split_name = re.split(comma_and_spaces, name)
            first_name = split_name[1].split()
            for i, w in enumerate(first_name):
                first_name[i] = Cleaner(w).clean_title()
            new_first_name = ' '.join(first_name)
            surname = split_name[0].split()
            for i, w in enumerate(surname):
                surname[i] = Cleaner(w).clean_title()
            new_surname = ' '.join(surname)
            if new_surname:
                new_name = new_surname + ', ' + new_first_name
            else:
                new_name = ''
        else:
            split_name = name.split()
            for i, w in enumerate(split_name):
                split_name[i] = Cleaner(w).clean_title()
            new_name = ' '.join(split_name)
        return new_name
        
    def remove_unwanted_characters(self) -> str:
        '''
        This method helps remove unwanted characters from authors' names. 
        Such characters are all characters other than letters, numbers, space, '&', apostroph, or dots that are not preceded by letters. 
        Numbers and '&' are significant if the author is an organization and not a person.
        Finally, hyphens are normalized, Unicode encodings decoded, and extra spaces removed.

        :returns: str -- The cleaned name
        '''
        unwanted_characters = {'[', ']', '{', '}', '(', ')', '?', ';', ','}
        clean_string = str()
        for i, c in enumerate(self.string):
            if c == '.':
                if self.string[i-1].isalpha():
                    clean_string += c
            elif c not in unwanted_characters:
                clean_string += c
        clean_string = ' '.join(clean_string.split()).strip()
        clean_string = html.unescape(clean_string)
        clean_string = Cleaner(clean_string).normalize_hyphens()
        return clean_string
    
    @staticmethod
    def clean_ra_list(ra_list:list) -> list:
        '''
        This method removes responsible agents reported as 'Not Available'.

        :returns: list -- The cleaned responsible agents' list
        '''
        new_ra_list = list()
        for ra in ra_list:
            if ',' in ra:
                split_name = re.split(comma_and_spaces, ra)
                first_name = split_name[1] if split_name[1].lower() != 'not available' else ''
                given_name = split_name[0] if split_name[0].lower() != 'not available' else ''
                if given_name:
                    if first_name:
                        new_ra_list.append(ra)
                    else:
                        new_ra_list.append(f'{given_name}, ')
                else:
                    continue
            else:
                if ra.lower() != 'not available':
                    new_ra_list.append(ra)
        return new_ra_list
        
    
    def normalize_id(self, valid_dois_cache:CSVManager=None) -> Union[str, None]:
        '''
        This function verifies and normalizes identifiers whose schema corresponds to a DOI, an ISSN, an ISBN or an ORCID.

        :returns: Union[str, None] -- The normalized identifier if it is valid, None otherwise
        '''
        identifier = self.string.split(':', 1)
        schema = identifier[0].lower()
        value = identifier[1]
        if schema == 'doi':
            use_api_service = True if valid_dois_cache is not None else False
            doi_manager = DOIManager(valid_doi=valid_dois_cache, use_api_service=use_api_service)
            valid_id = doi_manager.normalise(value, include_prefix=True) if doi_manager.is_valid(value) else None
        elif schema == 'isbn':
            isbn_manager = ISBNManager()
            valid_id = isbn_manager.normalise(value, include_prefix=True) if isbn_manager.is_valid(value) else None
        elif schema == 'issn':
            issn_manager = ISSNManager()
            valid_id = issn_manager.normalise(value, include_prefix=True) if issn_manager.is_valid(value) else None
        elif schema == 'orcid':
            orcid_manager = ORCIDManager()
            valid_id = orcid_manager.normalise(value, include_prefix=True) if orcid_manager.is_valid(value) else None
        else:
            valid_id = f'{schema}:{value}'
        return valid_id

    @classmethod
    def clean_volume_and_issue(cls, row:dict) -> None:
        output = {'volume': '', 'issue': '', 'pub_date': ''}
        for field in {'volume', 'issue'}:
            vi = row[field]
            vi = Cleaner(vi).normalize_hyphens()
            vi = Cleaner(vi).normalize_spaces().strip()
            vi = html.unescape(vi)
            for pattern, strategy in invalid_vi_patterns.items():
                pattern = f'^{pattern}$'
                capturing_groups = re.search(pattern, vi, re.IGNORECASE)
                if capturing_groups:
                    if strategy == 'del':
                        row[field] = ''
                    elif strategy == 'do_nothing':
                        row[field] = vi
                    elif strategy == 's)':
                        row[field] = f'{vi}s)'
                    else:
                        row[field] = ''
                        whatever, volume, issue, pub_date = cls.fix_invalid_vi(capturing_groups, strategy)
                        row[field] = whatever if whatever else row[field]
                        output['volume'] = volume if volume else ''
                        output['issue'] = issue if issue else ''
                        output['pub_date'] = pub_date if pub_date else ''
        row['volume'] = output['volume'] if not row['volume'] else row['volume']
        row['issue'] = output['issue'] if not row['issue'] else row['issue']
        row['pub_date'] = output['pub_date'] if not row['pub_date'] else row['pub_date']
        switch_vi = {'volume': '', 'issue': ''}
        for field in {'volume', 'issue'}:
            vi = row[field]
            for pattern in volumes_valid_patterns:
                pattern = f'^{pattern}$'
                if re.search(pattern, vi, re.IGNORECASE):
                    if field == 'issue':
                        switch_vi['volume'] = vi
            for pattern in issues_valid_patterns:
                pattern = f'^{pattern}$'
                if re.search(pattern, vi, re.IGNORECASE):
                    if field == 'volume':
                        switch_vi['issue'] = vi
        if switch_vi['volume'] and switch_vi['issue']:
            row['volume'] = switch_vi['volume']
            row['issue'] = switch_vi['issue']
        elif switch_vi['volume'] and not row['volume']:
            row['volume'] = switch_vi['volume']
            row['issue'] = ''
            row['type'] = 'journal volume' if row['type'] == 'journal issue' else row['type']
        elif switch_vi['issue'] and not row['issue']:
            row['issue'] = switch_vi['issue']
            row['volume'] = ''
            row['type'] = 'journal issue' if row['type'] == 'journal volume' else row['type']
    
    @staticmethod
    def fix_invalid_vi(capturing_groups:re.Match, strategy:str) -> Tuple[str, str, str, str]:
        vol_group = 1 if 'vol_iss' in strategy else 2
        iss_group = 1 if 'iss_vol' in strategy else 2
        whatever = None
        volume = None
        issue = None
        pub_date = None
        if 'vol' in strategy and 'iss' in strategy:
            volume = capturing_groups.group(vol_group)
            issue = capturing_groups.group(iss_group)
            if 'year' in strategy:
                pub_date = capturing_groups.group(3)
        elif strategy == 'all':
            whatever = capturing_groups.group(1)
        elif strategy == 'sep':
            first = capturing_groups.group(1)
            second = capturing_groups.group(2)
            whatever = f'{first}-{second}'
        return whatever, volume, issue, pub_date

    def remove_ascii(self):
        unwanted_chars = '\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff'
        clean_string = ''.join([' ' if c in unwanted_chars else c for c in self.string])
        clean_string = ' '.join(clean_string.split())
        return clean_string           


# -*- coding: utf-8 -*-
# Copyright (c) 2021, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from requests import get
from os.path import exists
from csv import DictReader, DictWriter
from json import loads
from time import sleep


MAX_TRY = 5
SLEEPING_TIME = 5
csv_headers = (
    "id", "name", "prefix"
)
headers = {
    "User-Agent": 
    "OpenCitations "
    "(http://opencitations.net; mailto:contact@opencitations.net)"
}


def get_via_requests(get_url):
    tentative = 0
    print("")

    while tentative < MAX_TRY:
        tentative += 1

        try:
            print("Getting data from", get_url, "- tentative:", tentative)
            r = get(get_url, headers=headers, timeout=10)
            if r.status_code == 200:
                print("\tdata downloaded")
                r.encoding = "utf-8"

                return loads(r.text)
            elif r.status_code == 404:
                return None
            else:
                print("\tdata not downloaded, trying again in ", SLEEPING_TIME, "seconds - status:", r.status_code)
                sleep(SLEEPING_TIME)
        except Exception as e:
            print("\tdata not downloaded, trying again in ", SLEEPING_TIME, "seconds - exception:", e.message)
            sleep(SLEEPING_TIME)


def get_publishers(offset):
    get_url = "https://api.crossref.org/members?rows=1000&offset=" + str(offset)
    req = get_via_requests(get_url)
    
    if req is not None:
        r_json = req.get("message")
        if r_json is not None:
            offset += 1000
            print("\tnext offset is", offset)
            total_results = int(r_json.get("total-results"))
            items = r_json.get("items")

            return items, offset, total_results
                

def process(out_path):
    pub_ids = set()

    if exists(out_path):
        with open(out_path, encoding="utf8") as f:
            csv_reader = DictReader(f, csv_headers)
            for row in csv_reader:
                pub_ids.add(row["id"])

    # cursor = "*"
    offset = 0
    tot = 10000000000
    pub_count = 0
    while offset < tot:
        result, offset, tot = get_publishers(offset)

        if result is not None:
            for publisher in result:
                pub_count += 1
                cur_id = str(publisher["id"])
                if cur_id not in pub_ids:
                    pub_ids.add(cur_id)
                    cur_name = publisher["primary-name"]
                    prefixes = set()
                    for prefix in publisher["prefix"]:
                        prefix_value = prefix["value"]
                        if prefix_value not in prefixes:
                            prefixes.add(prefix_value)
                            store_csv_on_file(out_path, csv_headers, {
                                "id": cur_id, "name": cur_name, "prefix": prefix_value})
    
    if pub_count == tot:
        print("\n\nAll publishers correctly downloaded")
    else:
        print("\n\nOnly %s of the total %s publishers "
              "have been downloaded" % (pub_count, tot))

def store_csv_on_file(f_path, header, json_obj):
    f_exists = exists(f_path)
    with open(f_path, "a", encoding="utf8", newline='') as f:
        dw = DictWriter(f, header)
        if not f_exists:
            dw.writeheader()
        dw.writerow(json_obj)



from oc_meta.lib.file_manager import write_csv, pathoo
from tqdm import tqdm
from zipfile import ZipFile
import csv
import io
import os


def get_ids_from_citations(citations_dir:str, output_dir:str, threshold:int=10000, verbose:bool=False) -> None:
    '''
    This script extracts the identifiers of the citing and cited documents from citation data organized in the CSV format accepted by OpenCitations.

    :params citations_dir: the directory containing the citations files, either in CSV or ZIP format
    :type citations_dir: str
    :params output_dir: directory of the output CSV files
    :type output_dir: str
    :params verbose: show a loading bar, elapsed time and estimated time
    :type verbose: bool
    :returns: None
    '''
    if not any(file.endswith('.csv') or file.endswith('.zip') for  _, _, files in os.walk(citations_dir) for file in files):
        raise RuntimeError('I did not find CSV or ZIP files in the given directory')
    ids_found = set()
    if os.path.isdir(output_dir):
        if verbose:
            print(f'[INFO: get_ids_from_citations_data] Looking for previously stored IDs')
        for filename in os.listdir(output_dir):
            with open(os.path.join(output_dir, filename), 'r', encoding='utf8', newline='') as f:
                reader = csv.DictReader(f)
                for row in reader:
                    ids_found.add(row['id'])
        if verbose:
            print(f'[INFO: get_ids_from_citations] {len(ids_found)} IDs found')
    else:
        pathoo(output_dir)
    if verbose:
        pbar = tqdm(total=get_files_count(citations_dir))
    file_counter = 1
    output_csv = list()
    for fold, _, files in os.walk(citations_dir):
        for file in files:
            cur_file = file_counter + len(ids_found)
            if file.endswith('.csv'):
                data = csv.DictReader(open(os.path.join(fold, file), 'r', encoding='utf8'))
                process_data(data, ids_found, output_csv)
                if file_counter % threshold == 0:
                    write_csv(path=os.path.join(output_dir, f'{cur_file-threshold+1}-{cur_file}.csv'), datalist=output_csv)
                    output_csv = list()
                if verbose:
                    pbar.update()
                file_counter += 1
            elif file.endswith('.zip'):
                with ZipFile(os.path.join(citations_dir, file), 'r') as archive:
                    for name in archive.namelist():
                        cur_file = file_counter + len(ids_found)
                        with archive.open(name) as infile:
                            data = csv.DictReader(io.TextIOWrapper(infile, 'utf-8'))
                            process_data(data, ids_found, output_csv)
                            if file_counter % threshold == 0:
                                write_csv(path=os.path.join(output_dir, f'{cur_file-threshold+1}-{cur_file}.csv'), datalist=output_csv)
                                output_csv = list()
                            if verbose:
                                pbar.update()
                            file_counter += 1
    if output_csv:
        write_csv(path=os.path.join(output_dir, f'{cur_file + 1 - (cur_file % threshold)}-{cur_file}.csv'), datalist=output_csv)
    if verbose:
        pbar.close()
        print(f'[INFO: get_ids_from_citations] {len(ids_found)} IDs stored')

def process_data(data:csv.DictReader, ids_found:set, output_csv:list) -> None:
    for row in data:
        citing = row['citing']
        cited = row['cited']
        if citing not in ids_found:
            output_csv.append({'id':row['citing']})
            ids_found.add(citing)
        if cited not in ids_found:
            output_csv.append({'id':row['cited']})
            ids_found.add(cited)

def get_files_count(citations_dir:str) -> int:
    if any(file.endswith('.csv') for  _, _, files in os.walk(citations_dir) for file in files):
        file_count = sum(len(files) for _, _, files in os.walk(citations_dir))
    elif any(file.endswith('.zip') for  _, _, files in os.walk(citations_dir) for file in files):
        file_count = 0
        for _, _, files in os.walk(citations_dir):
            for file in files:
                with ZipFile(os.path.join(citations_dir, file), 'r') as archive:
                    file_count += len(archive.namelist())
    return file_count




from argparse import ArgumentParser
from oc_meta.plugins.get_ids_from_citations import get_ids_from_citations


if __name__ == '__main__':
    arg_parser = ArgumentParser('run_preprocess.py', description='This script extracts the identifiers of the citing and cited documents from citation data organized in the CSV format accepted by OpenCitations. ')
    arg_parser.add_argument('-c', '--citations', dest='citations_dir', required=True,
                            help='The directory containing the citations files, either in CSV or ZIP format')
    arg_parser.add_argument('-out', '--output', dest='output_dir', required=True,
                            help='Directory of the output CSV files')
    arg_parser.add_argument('-t', '--threshold', dest='threshold', required=False,
                            help='Number of files to save after')
    arg_parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', required=False,
                            help='Show a loading bar, elapsed time and estimated time')
    args = arg_parser.parse_args()
    get_ids_from_citations(citations_dir=args.citations_dir, output_dir=args.output_dir, threshold=args.threshold, verbose=args.verbose)


import csv
import os
import sys
from contextlib import contextmanager
from oc_meta.lib.cleaner import Cleaner
from typing import List, Dict, Set



def get_data(filepath:str) -> List[Dict[str, str]]:
    field_size_changed = False
    cur_field_size = 128
    data = list()
    while not data:
        try:
            with open(filepath, 'r', encoding='utf8') as data_initial:
                valid_data = (Cleaner(line).normalize_spaces().replace('\0','') for line in data_initial)
                data = list(csv.DictReader(valid_data, delimiter=','))
        except csv.Error:
            cur_field_size *= 2
            csv.field_size_limit(cur_field_size)
            field_size_changed = True
    if field_size_changed:
        csv.field_size_limit(128)
    return data

def pathoo(path):
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))

def write_csv(path:str, datalist:List[dict], fieldnames:list=None) -> None:
    if not datalist:
        fieldnames = []
    else:
        fieldnames = datalist[0].keys() if fieldnames is None else fieldnames
    pathoo(path)
    with open(path, 'w', newline='', encoding='utf-8') as output_file:
        dict_writer = csv.DictWriter(f=output_file, fieldnames=fieldnames, delimiter=',', quotechar='"', quoting=csv.QUOTE_NONNUMERIC)
        dict_writer.writeheader()
        dict_writer.writerows(datalist)

def normalize_path(path:str) -> str:
    normal_path = path.replace('\\', '/').replace('/', os.sep)
    return normal_path

def init_cache(cache_filepath:str) -> Set[str]:
    completed = set()
    if cache_filepath:
        if not os.path.exists(cache_filepath):
            pathoo(cache_filepath)
        else:
            with open(cache_filepath, 'r', encoding='utf-8') as cache_file:
                completed = {line.rstrip('\n') for line in cache_file}
    return completed

@contextmanager
def suppress_stdout():
    with open(os.devnull, 'w') as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:  
            yield
        finally:
            sys.stdout = old_stdout

def sort_files(files_to_be_processed:list) -> None:
    if all(filename.replace('.csv', '').isdigit() for filename in files_to_be_processed):
        files_to_be_processed = sorted(files_to_be_processed, key=lambda filename: int(filename.replace('.csv', '')))
    elif all(filename.split('_')[-1].replace('.csv', '').isdigit() for filename in files_to_be_processed):
        files_to_be_processed = sorted(files_to_be_processed, key=lambda filename: int(filename.split('_')[-1].replace('.csv', '')))
    return files_to_be_processed



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2019, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


from oc_meta.lib.id_manager.identifiermanager import IdentifierManager
from re import sub
from urllib.parse import unquote, quote
from requests import get
from json import loads
from oc_meta.lib.csvmanager import CSVManager
from requests import ReadTimeout


class DOIManager(IdentifierManager):
    def __init__(self, valid_doi=None, use_api_service=True):
        if valid_doi is None:
            valid_doi = CSVManager()

        self.api = "https://doi.org/api/handles/"
        self.valid_doi = valid_doi
        self.use_api_service = use_api_service
        self.p = "doi:"
        super(DOIManager, self).__init__()

    def set_valid(self, id_string):
        doi = self.normalise(id_string, include_prefix=True)

        if self.valid_doi.get_value(doi) is None:
            self.valid_doi.add_value(doi, "v")

    def is_valid(self, id_string):
        doi = self.normalise(id_string, include_prefix=False)
        if doi is None:
            return False
        elif self.use_api_service:
            if self.valid_doi.get_value(doi) is None:
                if self.__doi_exists(doi):
                    self.valid_doi.add_value(doi, "v")
                else:
                    self.valid_doi.add_value(doi, "i")
            return "v" in self.valid_doi.get_value(doi)
        elif doi and not self.use_api_service:
            return True

    def normalise(self, id_string, include_prefix=False):
        try:
            doi_string = sub("\0+", "", sub("\s+", "", unquote(id_string[id_string.index("10."):])))
            return "%s%s" % (self.p if include_prefix else "", doi_string.lower().strip())
        except:  # Any error in processing the DOI will return None
            return None

    def __doi_exists(self, doi):
        tentative = 3
        while tentative:
            tentative -= 1
            try:
                r = get(self.api + quote(doi), headers=self.headers, timeout=30)
                if r.status_code == 200:
                    r.encoding = "utf-8"
                    json_res = loads(r.text)
                    return json_res.get("responseCode") == 1
            except ReadTimeout:
                pass  # Do nothing, just try again
        return False



import os
import re
import yaml
from oc_meta.lib.master_of_regex import ids_inside_square_brackets
from oc_meta.lib.finder import ResourceFinder
from oc_meta.lib.file_manager import write_csv, normalize_path
from typing import Tuple
from tqdm import tqdm


class CSVGenerator:
    def __init__(self, config:str):
        with open(config, encoding='utf-8') as file:
            settings = yaml.full_load(file)
        triplestore_url = settings['triplestore_url']
        self.info_dir = normalize_path(settings['info_dir'])
        self.base_iri = settings['base_iri'][:-1] if settings['base_iri'][-1] == '/' else settings['base_iri']
        self.supplier_prefix = settings['supplier_prefix']
        self.dir_split_number = settings['dir_split_number']
        self.output_csv_dir = os.path.join(normalize_path(settings['output_csv_dir']), str(self.supplier_prefix))
        self.items_per_file = settings['items_per_file']
        self.verbose = settings['verbose']
        self.forbidden_types = {'journal issue', 'journal volume', 'journal', 'book set', 'book series', 'book part', 'book section'}
        self.finder = ResourceFinder(triplestore_url, self.base_iri)        
    
    def generate_csv(self) -> None:
        '''
        This method generates CSVs from the Meta triplestore.
        '''
        counter = 1
        br_info_dir_path = os.path.join(self.info_dir, 'creator', f'info_file_br.txt')
        number_of_entities = self.__read_number(br_info_dir_path)
        last_file, dir_number = self.__skip_files()
        counter += last_file
        self.cur_output_dir = os.path.join(self.output_csv_dir, str(dir_number))
        if self.verbose:
            pbar = tqdm(total=number_of_entities)
            pbar.update(last_file)
        self.data = list()
        while counter <= number_of_entities:
            row = dict()
            metaid = self.supplier_prefix + str(counter)
            res = f'{self.base_iri}/br/{metaid}'
            res_exists_and_is_relevant = self.finder.check_type(res=res, forbidden_types=self.forbidden_types)
            if res_exists_and_is_relevant:
                br_info = self.finder.retrieve_br_info_from_meta(metaid)
                br_title_and_ids = self.finder.retrieve_br_from_meta(metaid)
                row.update(br_info)
                ids = [id[1] for id in br_title_and_ids[1]]
                row['id'] = f'meta:br/{metaid} ' + ' '.join(ids)
                row['title'] = br_title_and_ids[0]
                row['author'] = self.__get_resp_agents(metaid, 'author')
                row['editor'] = self.__get_resp_agents(metaid, 'editor')
                row['page'] = br_info['page'][1] if row['page'] else ''
                row['venue'] = self.__get_venue(br_info)
                row['publisher'] = self.finder.retrieve_publisher_from_br_metaid(metaid)
                row = {k:v if v else '' for k,v in row.items()}
                self.data.append(row)
            if self.data:
                self.__store_csv(counter)
            counter += 1
            if self.verbose:
                pbar.update()
        if self.verbose:
            pbar.close()
    
    def __get_resp_agents(self, metaid:str, column:str) -> str:
        resp_agents = self.finder.retrieve_ra_sequence_from_br_meta(metaid, column)
        output = ''
        if resp_agents:
            full_resp_agents = list()
            for item in resp_agents:
                for _, resp_agent in item.items():
                    author_name = resp_agent[0]
                    ids = [f'meta:ra/{resp_agent[2]}']
                    ids.extend([id[1] for id in resp_agent[1]])
                    author_ids = '[' + ' '.join(ids) + ']'
                    full_resp_agent =  author_name + ' ' + author_ids
                    full_resp_agents.append(full_resp_agent)
            output = '; '.join(full_resp_agents)
        return output
    
    def __get_venue(self, br_info:dict):
        venue_field = ''
        if br_info['venue']:
            venue_metaid = re.search(ids_inside_square_brackets, br_info['venue']).group(1).replace('meta:br/', '')
            venue_title_and_ids = self.finder.retrieve_br_from_meta(venue_metaid)
            venue_title = venue_title_and_ids[0]
            venue_ids = [f'meta:br/{venue_metaid}']
            venue_ids.extend([id[1] for id in venue_title_and_ids[1]])
            venue_ids_string = '[' + ' '.join(venue_ids) + ']'
            venue_field = venue_title + ' ' + venue_ids_string
        return venue_field
    
    def __store_csv(self, counter:int) -> None:
        if counter != 0 and counter % self.items_per_file == 0:
            if os.path.exists(self.cur_output_dir):
                if len(os.listdir(self.cur_output_dir)) % self.dir_split_number == 0:
                    cur_dir = str(int(counter - self.items_per_file + self.dir_split_number * self.items_per_file))
                    self.cur_output_dir = os.path.join(self.output_csv_dir, cur_dir)
            path = os.path.join(self.cur_output_dir, f'{counter}.csv')
            fieldnames = ['id', 'title', 'author', 'pub_date', 'venue', 'volume', 'issue', 'page', 'type', 'publisher', 'editor']
            write_csv(path=path, datalist=self.data, fieldnames=fieldnames)
            self.data = list()
    
    def __skip_files(self) -> Tuple[int, int]:
        last_file = 0
        dir_number = self.dir_split_number * self.items_per_file
        if os.path.exists(self.output_csv_dir):
            folders = os.listdir(self.output_csv_dir)
            if folders:
                last_folder = sorted(folders, key=lambda x: int(x))[-1]
                files = os.listdir(os.path.join(self.output_csv_dir, last_folder))
                if files:
                    last_file = sorted(files, key=lambda x: int(x.replace('.csv', '')))[-1]
                    last_file = int(last_file.replace('.csv', ''))
                    dir_number = dir_number * len(folders)
        return last_file, dir_number

    @staticmethod
    def __read_number(file_path: str) -> Tuple[int, int]:
        with open(file_path, "rb") as f:
            number = int(f.readlines()[0].decode('ascii').strip())
            return number



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


from oc_meta.lib.finder import ResourceFinder
from oc_meta.lib.master_of_regex import comma_and_spaces, name_and_ids, semicolon_in_people_field
from oc_meta.scripts.creator import Creator
from oc_ocdm.graph import GraphSet
from rdflib import URIRef
import re


class RespAgentsCreator(Creator):
    def __init__(self, data:list, endpoint:str, base_iri:str, info_dir:str, supplier_prefix:str, resp_agent:str, ra_index:dict, preexisting_entities:set):
        self.url = base_iri
        self.setgraph = GraphSet(self.url, info_dir, supplier_prefix, wanted_label=False)
        self.finder = ResourceFinder(ts_url = endpoint, base_iri = base_iri)
        self.resp_agent = resp_agent
        self.ra_id_schemas = {'crossref', 'orcid', 'viaf', 'wikidata'}
        self.br_id_schemas = {'doi', 'issn', 'isbn', 'pmid', 'pmcid', 'url', 'wikidata', 'wikipedia'}
        self.schemas = self.ra_id_schemas.union(self.br_id_schemas)
        self.ra_index = self.indexer_id(ra_index)
        self.preexisting_entities = preexisting_entities
        self.preexisting_graphs = dict()
        self.data = data

    def creator(self, source=None):
        self.src = source
        for row in self.data:
            authors = row['author']
            publisher = row['publisher']
            editor = row['editor']
            self.author_action(authors)
            if publisher:
                self.publisher_action(publisher)
            if editor:
                self.editor_action(editor)
        return self.setgraph

    def author_action(self, authors):
        if authors:
            authorslist = re.split(semicolon_in_people_field, authors)
            for aut in authorslist:
                aut_and_ids = re.search(name_and_ids, aut)
                aut_id = aut_and_ids.group(2)
                aut_id_list = aut_id.split()
                for identifier in aut_id_list:
                    if 'meta:' in identifier:
                        identifier = str(identifier).replace('meta:', '')
                        preexisting_entity = True if identifier in self.preexisting_entities else False
                        url = URIRef(self.url + identifier)
                        preexisting_graph = self.finder.get_preexisting_graph(url, self.preexisting_graphs) if preexisting_entity else None
                        pub_aut = self.setgraph.add_ra(self.resp_agent, source=self.src, res=url, preexisting_graph=preexisting_graph)
                        author_name = aut_and_ids.group(1)
                        if ',' in author_name:
                            author_name_splitted = re.split(comma_and_spaces, author_name)
                            first_name = author_name_splitted[1]
                            last_name = author_name_splitted[0]
                            if first_name.strip():
                                pub_aut.has_given_name(first_name)
                            pub_aut.has_family_name(last_name)
                        else:
                            pub_aut.has_name(author_name)
                # lists of authors' IDs
                for identifier in aut_id_list:
                    self.id_creator(pub_aut, identifier, ra=True)

    def publisher_action(self, publisher):
        publ_and_ids = re.search(name_and_ids, publisher)
        publ_id = publ_and_ids.group(2)
        publ_id_list = publ_id.split()
        for identifier in publ_id_list:
            if 'meta:' in identifier:
                identifier = str(identifier).replace('meta:', '')
                preexisting_entity = True if identifier in self.preexisting_entities else False
                url = URIRef(self.url + identifier)
                publ_name = publ_and_ids.group(1)
                preexisting_graph = self.finder.get_preexisting_graph(url, self.preexisting_graphs) if preexisting_entity else None
                publ = self.setgraph.add_ra(self.resp_agent, source=self.src, res=url, preexisting_graph=preexisting_graph)
                publ.has_name(publ_name)
        for identifier in publ_id_list:
            self.id_creator(publ, identifier, ra=True)

    def editor_action(self, editor):
        editorslist = re.split(semicolon_in_people_field, editor)
        for ed in editorslist:
            ed_and_ids = re.search(name_and_ids, ed)
            ed_id = ed_and_ids.group(2)
            ed_id_list = ed_id.split(' ')
            for identifier in ed_id_list:
                if 'meta:' in identifier:
                    identifier = str(identifier).replace('meta:', '')
                    preexisting_entity = True if identifier in self.preexisting_entities else False
                    url = URIRef(self.url + identifier)
                    preexisting_graph = self.finder.get_preexisting_graph(url, self.preexisting_graphs) if preexisting_entity else None
                    pub_ed = self.setgraph.add_ra(self.resp_agent, source=self.src, res=url, preexisting_graph=preexisting_graph)
                    editor_name = ed_and_ids.group(1)
                    if ',' in editor_name:
                        editor_name_splitted = re.split(comma_and_spaces, editor_name)
                        firstName = editor_name_splitted[1]
                        lastName = editor_name_splitted[0]
                        if firstName.strip():
                            pub_ed.has_given_name(firstName)
                        pub_ed.has_family_name(lastName)
                    else:
                        pub_ed.has_name(editor_name)
            # lists of editor's IDs
            for identifier in ed_id_list:
                self.id_creator(pub_ed, identifier, ra=True)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2019 Silvio Peroni <essepuntato@gmail.com>
# Copyright 2019-2020 Fabio Mariani <fabio.mariani555@gmail.com>
# Copyright 2021 Simone Persiani <iosonopersia@gmail.com>
# Copyright 2021-2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


from argparse import ArgumentParser
from datetime import datetime
from filelock import FileLock
from itertools import cycle
from oc_meta.lib.csvmanager import CSVManager
from oc_meta.lib.file_manager import get_data, normalize_path, pathoo, suppress_stdout, init_cache, sort_files
from oc_meta.plugins.multiprocess.resp_agents_creator import RespAgentsCreator
from oc_meta.plugins.multiprocess.resp_agents_curator import RespAgentsCurator
from oc_meta.scripts.creator import Creator
from oc_meta.scripts.curator import Curator
from oc_ocdm import Storer
from oc_ocdm.prov import ProvSet
from pathlib import Path
from pebble import ProcessPool, ProcessFuture
from shutil import make_archive
from sys import platform
from time_agnostic_library.support import generate_config_file
from typing import List, Set, Tuple
import csv
import os
import yaml


class MetaProcess:
    def __init__(self, config:str):
        with open(config, encoding='utf-8') as file:
            settings = yaml.full_load(file)
        self.config = config
        # Mandatory settings
        self.triplestore_url = settings['triplestore_url']
        self.input_csv_dir = normalize_path(settings['input_csv_dir'])
        self.base_output_dir = normalize_path(settings['base_output_dir'])
        self.resp_agent = settings['resp_agent']
        self.info_dir = os.path.join(self.base_output_dir, 'info_dir')
        self.output_csv_dir = os.path.join(self.base_output_dir, 'csv')
        self.output_rdf_dir = os.path.join(self.base_output_dir, f'rdf{os.sep}')
        self.indexes_dir = os.path.join(self.base_output_dir, 'indexes')
        self.cache_path = os.path.join(self.base_output_dir, 'cache.txt')
        self.errors_path = os.path.join(self.base_output_dir, 'errors.txt')
        # Optional settings
        self.base_iri = settings['base_iri']
        self.context_path = settings['context_path']
        self.dir_split_number = settings['dir_split_number']
        self.items_per_file = settings['items_per_file']
        self.default_dir = settings['default_dir']
        self.rdf_output_in_chunks = settings['rdf_output_in_chunks']
        self.source = settings['source']
        self.valid_dois_cache = CSVManager() if bool(settings['use_doi_api_service']) == True else None
        self.workers_number = int(settings['workers_number'])
        supplier_prefix:str = settings['supplier_prefix']
        self.supplier_prefix = supplier_prefix[:-1] if supplier_prefix.endswith('0') else supplier_prefix
        self.verbose = settings['verbose']
        # Time-Agnostic_library integration
        self.time_agnostic_library_config = os.path.join(os.path.dirname(config), 'time_agnostic_library_config.json')
        if not os.path.exists(self.time_agnostic_library_config):
            generate_config_file(config_path=self.time_agnostic_library_config, dataset_urls=[self.triplestore_url], dataset_dirs=list(),
                provenance_urls=settings['provenance_endpoints'], provenance_dirs=list(), 
                blazegraph_full_text_search=settings['blazegraph_full_text_search'], graphdb_connector_name=settings['graphdb_connector_name'], 
                cache_endpoint=settings['cache_endpoint'], cache_update_endpoint=settings['cache_update_endpoint'])

    def prepare_folders(self) -> Set[str]:
        completed = init_cache(self.cache_path)
        files_in_input_csv_dir = {filename for filename in os.listdir(self.input_csv_dir) if filename.endswith('.csv')}
        files_to_be_processed = sort_files(files_in_input_csv_dir.difference(completed))
        for dir in [self.output_csv_dir, self.indexes_dir, self.output_rdf_dir]:
            pathoo(dir)
        if self.rdf_output_in_chunks:
            data_dir = os.path.join(self.output_rdf_dir, 'data' + os.sep)
            prov_dir = os.path.join(self.output_rdf_dir, 'prov' + os.sep)
            for dir in [prov_dir, data_dir]:
                pathoo(dir)
        csv.field_size_limit(128)
        return files_to_be_processed

    def curate_and_create(self, filename:str, cache_path:str, errors_path:str, worker_number:int=None, resp_agents_only:bool=False) -> None:
        if os.path.exists(os.path.join(self.base_output_dir, '.stop')):
            return
        try:
            filepath = os.path.join(self.input_csv_dir, filename)
            data = get_data(filepath)
            supplier_prefix = f'{self.supplier_prefix}0' if worker_number is None else f'{self.supplier_prefix}{str(worker_number)}0'
            # Curator
            self.info_dir = os.path.join(self.info_dir, supplier_prefix) if worker_number else self.info_dir
            curator_info_dir = os.path.join(self.info_dir, 'curator' + os.sep)
            if resp_agents_only:
                curator_obj = RespAgentsCurator(data=data, ts=self.triplestore_url, prov_config=self.time_agnostic_library_config, info_dir=curator_info_dir, base_iri=self.base_iri, prefix=supplier_prefix)
            else:
                curator_obj = Curator(data=data, ts=self.triplestore_url, prov_config=self.time_agnostic_library_config, info_dir=curator_info_dir, base_iri=self.base_iri, prefix=supplier_prefix, valid_dois_cache=self.valid_dois_cache)
            name = f"{datetime.now().strftime('%Y-%m-%dT%H_%M_%S_%f')}_{supplier_prefix}"
            curator_obj.curator(filename=name, path_csv=self.output_csv_dir, path_index=self.indexes_dir)
            # Creator
            creator_info_dir = os.path.join(self.info_dir, 'creator' + os.sep)
            if resp_agents_only:
                creator_obj = RespAgentsCreator(
                    data=curator_obj.data, endpoint=self.triplestore_url, base_iri=self.base_iri, info_dir=creator_info_dir, supplier_prefix=supplier_prefix, resp_agent=self.resp_agent, ra_index=curator_obj.index_id_ra, preexisting_entities=curator_obj.preexisting_entities)
            else:
                creator_obj = Creator(
                    data=curator_obj.data, endpoint=self.triplestore_url, base_iri=self.base_iri, info_dir=creator_info_dir, supplier_prefix=supplier_prefix, resp_agent=self.resp_agent, ra_index=curator_obj.index_id_ra,
                    br_index=curator_obj.index_id_br, re_index_csv=curator_obj.re_index, ar_index_csv=curator_obj.ar_index, vi_index=curator_obj.VolIss, preexisting_entities=curator_obj.preexisting_entities)
            creator = creator_obj.creator(source=self.source)
            # Provenance
            prov = ProvSet(creator, self.base_iri, creator_info_dir, wanted_label=False)
            prov.generate_provenance()
            # Storer
            res_storer = Storer(creator, context_map={}, dir_split=self.dir_split_number, n_file_item=self.items_per_file, default_dir=self.default_dir, output_format='json-ld')
            prov_storer = Storer(prov, context_map={}, dir_split=self.dir_split_number, n_file_item=self.items_per_file, output_format='json-ld')
            with suppress_stdout():
                self.store_data_and_prov(res_storer, prov_storer, filename)
            return {'message': 'success'}, cache_path, errors_path, filename
        except Exception as e:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            return {'message': message}, cache_path, errors_path, filename
    
    def store_data_and_prov(self, res_storer:Storer, prov_storer:Storer, filename:str) -> None:
        if self.rdf_output_in_chunks:
            filename_without_csv = filename[:-4]
            f = os.path.join(self.output_rdf_dir, 'data', filename_without_csv + '.json')
            res_storer.store_graphs_in_file(f, self.context_path)
            res_storer.upload_all(self.triplestore_url, self.output_rdf_dir, batch_size=100)
            f_prov = os.path.join(self.output_rdf_dir, 'prov', filename_without_csv + '.json')
            prov_storer.store_graphs_in_file(f_prov, self.context_path)
        else:
            res_storer.store_all(base_dir=self.output_rdf_dir, base_iri=self.base_iri, context_path=self.context_path)
            prov_storer.store_all(self.output_rdf_dir, self.base_iri, self.context_path)
            res_storer.upload_all(triplestore_url=self.triplestore_url, base_dir=self.output_rdf_dir, batch_size=100)

def run_meta_process(meta_process:MetaProcess, resp_agents_only:bool=False) -> None:
    delete_lock_files(base_dir=meta_process.base_output_dir)
    files_to_be_processed = meta_process.prepare_folders()
    max_workers = meta_process.workers_number
    if max_workers == 0:
        workers = list(range(1, os.cpu_count()))
    elif max_workers == 1:
        workers = [None]
    else:
        multiples_of_ten = {i for i in range(1, max_workers+1) if int(i) % 10 == 0}
        workers = [i for i in range(1, max_workers+len(multiples_of_ten)+1) if i not in multiples_of_ten]
    is_unix = platform in {'linux', 'linux2', 'darwin'}
    generate_gentle_buttons(meta_process.base_output_dir, meta_process.config, is_unix)
    files_chunks = chunks(list(files_to_be_processed), 1000) if is_unix else [files_to_be_processed]
    for files_chunk in files_chunks:
        with ProcessPool(max_workers=max_workers, max_tasks=1) as executor:
            for file_to_be_processed, worker_number in zip(files_chunk, cycle(workers)):
                future:ProcessFuture = executor.schedule(
                    function=meta_process.curate_and_create, 
                    args=(file_to_be_processed, meta_process.cache_path, meta_process.errors_path, worker_number, resp_agents_only)) 
                future.add_done_callback(task_done) 
        delete_lock_files(base_dir=meta_process.base_output_dir)
        if is_unix and not os.path.exists(os.path.join(meta_process.base_output_dir, '.stop')):
            save_results(meta_process.base_output_dir)
    if os.path.exists(meta_process.cache_path) and not os.path.exists(os.path.join(meta_process.base_output_dir, '.stop')):
        os.rename(meta_process.cache_path, meta_process.cache_path.replace('.txt', f'_{datetime.now().strftime("%Y-%m-%dT%H_%M_%S_%f")}.txt'))
    if not is_unix:
        delete_lock_files(base_dir=meta_process.base_output_dir)

def task_done(task_output:ProcessFuture) -> None:
    message, cache_path, errors_path, filename = task_output.result()
    if message['message'] == 'success':
        if not os.path.exists(cache_path):
            with open(cache_path, 'w', encoding='utf-8') as aux_file:
                aux_file.write(filename + '\n')
        else:
            with open(cache_path, 'r', encoding='utf-8') as aux_file:
                cache_data = aux_file.read().splitlines()
                cache_data.append(filename)
                data_sorted = sorted(cache_data, key=lambda filename: int(filename.replace('.csv', '')), reverse=False)
            with open(cache_path, 'w', encoding='utf-8') as aux_file:
                aux_file.write('\n'.join(data_sorted))
    else:
        with open(errors_path, 'a', encoding='utf-8') as aux_file:
            aux_file.write(f'{filename}: {message["message"]}' + '\n')

def chunks(lst:list, n:int) -> List[list]:
    '''Yield successive n-sized chunks from lst.'''
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def delete_lock_files(base_dir:list) -> None:
    for dirpath, _, filenames in os.walk(base_dir):
        for filename in filenames:
            if filename.endswith('.lock'):
                os.remove(os.path.join(dirpath, filename))

def save_results(output_dirpath:str) -> None:
    output_dirname = f"meta_output_{datetime.now().strftime('%Y-%m-%dT%H_%M_%S_%f')}"
    parent_dir = Path(output_dirpath)
    parent_dir = parent_dir.parent.absolute()
    make_archive(base_name=output_dirname, format='zip', base_dir=output_dirpath)

def generate_gentle_buttons(dir:str, config:str, is_unix:bool):
    if os.path.exists(os.path.join(dir, '.stop')):
        os.remove(os.path.join(dir, '.stop'))
    ext = 'sh' if is_unix else 'bat'
    with open (f'gently_run.{ext}', 'w') as rsh:
        rsh.write(f'python -m oc_meta.lib.stopper -t "{dir}" --remove\npython -m oc_meta.run.meta_process -c {config}')
    with open (f'gently_stop.{ext}', 'w') as rsh:
        rsh.write(f'python -m oc_meta.lib.stopper -t "{dir}" --add')


if __name__ == '__main__':
    arg_parser = ArgumentParser('meta_process.py', description='This script runs the OCMeta data processing workflow')
    arg_parser.add_argument('-c', '--config', dest='config', required=True, help='Configuration file directory')
    args = arg_parser.parse_args()
    meta_process = MetaProcess(config=args.config)
    run_meta_process(meta_process=meta_process)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2019, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from abc import ABC, abstractmethod 


class IdentifierManager(ABC):
    """This is the interface that must be implemented by any identifier manager
    for a particular identifier scheme. It provides the signatures of the methods
    for checking the validity of an identifier and for normalising it."""

    def __init__(self, **params):
        for key in params:
            setattr(self, key, params[key])

        self.headers = {
            "User-Agent": "Identifier Manager / OpenCitations Indexes "
                          "(http://opencitations.net; mailto:contact@opencitations.net)"
        }

    @abstractmethod
    def is_valid(self, id_string):
        pass

    @abstractmethod
    def normalise(self, id_string, include_prefix=False):
        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE


from argparse import ArgumentParser
from oc_meta.lib.file_manager import normalize_path
from oc_meta.run.meta_process import MetaProcess, run_meta_process
from oc_meta.plugins.multiprocess.prepare_multiprocess import prepare_relevant_items, split_csvs_in_chunks
import os
import shutil
import yaml


if __name__ == '__main__':
    arg_parser = ArgumentParser('prepare_multiprocess.py', description='Venues, authors and editors are preprocessed not to create duplicates when running Meta in multi-process')
    arg_parser.add_argument('-c', '--config', dest='config', required=True, help='Configuration file path')
    args = arg_parser.parse_args()
    config = args.config
    with open(config, encoding='utf-8') as file:
        settings = yaml.full_load(file)
    csv_dir = normalize_path(settings['input_csv_dir'])
    items_per_file = settings['items_per_file']
    workers_numbers = settings['workers_number']
    verbose = settings['verbose']
    meta_process = MetaProcess(config)
    TMP_DIR = os.path.join(meta_process.base_output_dir, 'tmp')
    if not all(os.path.exists(os.path.join(TMP_DIR, directory)) for directory in ['venues', 'ids', 'publishers', 'people']):
        split_csvs_in_chunks(csv_dir=csv_dir, output_dir=TMP_DIR, chunk_size=1000, verbose=verbose)
        os.rename(csv_dir, csv_dir + '_old')
        os.mkdir(csv_dir)
        for file in os.listdir(TMP_DIR):
            shutil.move(os.path.join(TMP_DIR, file), csv_dir)
        prepare_relevant_items(csv_dir=csv_dir, output_dir=TMP_DIR, items_per_file=items_per_file, verbose=verbose)
    venues_dir = os.path.join(TMP_DIR, 'venues')
    if os.path.isdir(venues_dir):
        meta_process.input_csv_dir = venues_dir
        run_meta_process(meta_process=meta_process)
    publishers_dir = os.path.join(TMP_DIR, 'publishers')
    if os.path.isdir(publishers_dir):
        meta_process.input_csv_dir = publishers_dir
        run_meta_process(meta_process=meta_process, resp_agents_only=True)
    people_dir = os.path.join(TMP_DIR, 'people')
    if os.path.isdir(people_dir):
        meta_process.input_csv_dir = people_dir
        run_meta_process(meta_process=meta_process, resp_agents_only=True)
    meta_process.workers_number = 1
    ids_dir = os.path.join(TMP_DIR, 'ids')
    if os.path.isdir(ids_dir):
        meta_process.input_csv_dir = ids_dir
        run_meta_process(meta_process=meta_process)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

__author__ = 'essepuntato'
import argparse
import os


class Stopper(object):
    def __init__(self, target_dir):
        self.target_dir = target_dir
        self.stop_file = target_dir + os.sep + ".stop"

    def add(self):
        if self.can_proceed():
            if not os.path.exists(self.target_dir):
                os.makedirs(self.target_dir)
            open(self.stop_file, "w").close()

    def remove(self):
        if not self.can_proceed():
            os.remove(self.stop_file)

    def can_proceed(self):
        return not os.path.exists(self.stop_file)


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser("stopper.py")
    arg_parser.add_argument("-t", "--target-dir", dest="target_dir", required=True,
                            help="The configuration file to access the ORCID API.")
    arg_parser.add_argument("--add", dest="add", default=True, action="store_true",
                            help="It will add a stop marker to the target directory.")
    arg_parser.add_argument("--remove", dest="remove", default=False, action="store_true",
                            help="It will remove the block marker from the target directory.")
    args = arg_parser.parse_args()

    stopper = Stopper(args.target_dir)
    if args.remove:
        stopper.remove()
        print("Stop marker removed from to '%s'" % args.target_dir)
    elif args.add:
        stopper.add()
        print("Stop marker added to '%s'" % args.target_dir)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2019, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


from oc_meta.lib.id_manager.identifiermanager import IdentifierManager
from re import sub, match


class ORCIDManager(IdentifierManager):
    def __init__(self):
        self.p = "orcid:"
        super(ORCIDManager, self).__init__()

    def is_valid(self, id_string):
        orcid = self.normalise(id_string)
        return orcid is not None and \
               match("^([0-9]{4}-){3}[0-9]{3}[0-9X]$", orcid) and \
               ORCIDManager.__check_digit(orcid)

    def normalise(self, id_string, include_prefix=False):
        try:
            orcid_string = sub("[^X0-9]", "", id_string.upper())
            return "%s%s-%s-%s-%s" % (self.p if include_prefix else "",
                                      orcid_string[:4], orcid_string[4:8], orcid_string[8:12], orcid_string[12:16])
        except:  # Any error in processing the ISSN will return None
            return None

    @staticmethod
    def __check_digit(orcid):
        total = 0
        for d in sub("[^X0-9]", "", orcid.upper())[:-1]:
            i = 10 if d == "X" else int(d)
            total = (total + i) * 2
        reminder = total % 11
        result = (12 - reminder) % 11
        return (str(result) == orcid[-1]) or (result == 10 and orcid[-1] == "X")



from oc_meta.plugins.crossref.extract_crossref_publishers import process
from argparse import ArgumentParser



if __name__ == "__main__":
    arg_parser = ArgumentParser("Extract publisher information from Crossref")
    arg_parser.add_argument("-o", "--output", required=True,
                            help="The output CSV file where to store relevant information.")
    args = arg_parser.parse_args()

    print("Start process")
    process(args.output)
    print("Process finished")



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2019, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


from oc_meta.lib.id_manager.identifiermanager import IdentifierManager
from re import sub, match


class ISSNManager(IdentifierManager):
    def __init__(self):
        self.p = "issn:"
        super(ISSNManager, self).__init__()

    def is_valid(self, id_string):
        issn = self.normalise(id_string)
        return issn is not None and issn != '0000-0000' and match("^[0-9]{4}-[0-9]{3}[0-9X]$", issn) and ISSNManager.__check_digit(issn)

    def normalise(self, id_string, include_prefix=False):
        try:
            issn_string = sub("[^X0-9]", "", id_string.upper())
            return "%s%s-%s" % (self.p if include_prefix else "", issn_string[:4], issn_string[4:8])
        except:  # Any error in processing the ISSN will return None
            return None

    @staticmethod
    def __check_digit(issn):
        result_partial_sum = 0
        for i, n in zip(range(8, 1, -1), issn[:4] + issn[5:8]):
            result_partial_sum += i * int(n)
        reminder = result_partial_sum % 11
        reminder_sub = 11 - reminder
        correct_check_digit = \
            str(reminder_sub) == issn[8] or \
            (reminder == 0 and issn[8] == "0") or \
            (reminder_sub == 10 and issn[8] == "X")

        result_full_sum = 0
        for i, n in zip(range(8, 0, -1), issn[:4] + issn[5:]):
            result_full_sum += i * (10 if n == "X" else int(n))
        confirm_check_digit = result_full_sum % 11 == 0

        return correct_check_digit and confirm_check_digit


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


FORBIDDEN_IDS = {'issn:0000-0000'}
VENUES = {'archival-document', 'book', 'book-part', 'book-section', 'book-series', 'book-set', 'edited-book', 'journal', 'journal-volume', 'journal-issue', 'monograph', 'proceedings-series', 'proceedings', 'reference-book', 'report-series', 'standard-series'}
CONTAINER_EDITOR_TYPES = {'book chapter', 'book part', 'book section', 'book track', 'component', 'journal article', 'proceedings article', 'reference entry'}
